from datetime import datetime, timedelta
from logging import getLogger as get_logger

import requests
from django.utils.module_loading import import_string

log = get_logger('selectel')
now = datetime.now


# get, put_stream, remove, info, list
# info -- content-length
class Auth:
    THRESHOLD = 300
    AUTH_URL = 'https://auth.selcdn.ru/'

    def __init__(self, config, requests):
        self.config = config
        self.requests = requests
        self.token = ''
        self.storage_url = ''
        self.expires = now()

    @property
    def user(self):
        return self.config['USERNAME']

    @property
    def key(self):
        return self.config['PASSWORD']

    @property
    def container_name(self):
        return self.config['CONTAINER']

    def build_url(self, key):
        self.renew_if_need()
        return '{storage_url}/{name}/{key}'.format(
            storage_url=self.storage_url.rstrip('/'),
            name=self.container_name.strip('/'),
            key=key.lstrip('/'),
        )

    def is_expired(self):
        return (self.expires - now()).total_seconds() < self.THRESHOLD

    def renew_if_need(self):
        if self.is_expired() or not self.storage_url:
            self.authenticate()

    def update_expires(self, delta):
        self.expires = now() + timedelta(seconds=int(delta))

    def update_auth_token(self, new_token):
        self.requests.headers.update({'X-Auth-Token': new_token})
        self.token = new_token

    def authenticate(self):
        log.debug('Need to authenticate with %s:%s', self.user, self.key)
        resp = self.requests.get(self.AUTH_URL, headers={
            'X-Auth-User': self.user,
            'X-Auth-Key': self.key
        })
        if resp.status_code != 204:
            log.debug('Got an unexpected response from auth: %s', resp.content)
            raise Exception("Selectel: Unexpected status code: %s" %
                            resp.status_code)
        self.storage_url = resp.headers['X-Storage-Url']
        self.update_expires(resp.headers['X-Expire-Auth-Token'])
        self.update_auth_token(resp.headers['X-Auth-Token'])

    def perform_request(self, http_method, key,
                        raise_exception=False, **kwargs):
        self.renew_if_need()
        resp = getattr(self.requests, http_method)(
            self.build_url(key),
            **kwargs
        )  # type: requests.Response
        if raise_exception:
            resp.raise_for_status()
        return resp


class Container:
    def __init__(self, config):
        self.config = config
        self.requests = self.get_requests()  # type: requests.Session
        self.auth = Auth(self.config, self.requests)

    @property
    def name(self):
        return self.auth.container_name

    def get_requests(self):
        return import_string(self.config.get(
            'REQUESTS_FACTORY',
            'django_selectel_storage.utils.requests_factory'
        ))(self.config)

    def build_url(self, key):
        return self.auth.build_url(key)

    def perform_request(self, http_method, key,
                        raise_exception=False, **kwargs):
        return self.auth.perform_request(http_method, key,
                                         raise_exception=raise_exception,
                                         **kwargs)

    def open(self, key):
        return self.perform_request('get', key, raise_exception=True,
                                    stream=True).raw

    def save(self, key, content, metadata=None):
        self.perform_request('put', key, data=content, raise_exception=True)
        return key

    def delete(self, key):
        self.perform_request('delete', key, raise_exception=False)

    def size(self, key):
        try:
            resp = self.perform_request('head', key, raise_exception=True)
            return int(resp.headers['Content-Length'])
        except requests.exceptions.HTTPError:
            raise IOError('Failed to get file size of {}'.format(key))

    def exists(self, key):
        return self.perform_request('head', key).status_code == 200

    def list(self, key):
        return {
            x['name']: x for x in self.perform_request('get', '', params={
                'format': 'json',
                'prefix': key
            }).json()
        }



class SelectelException(ValueError):
    pass


class InvalidSchema(SelectelException):
    pass


class EmptyUsername(SelectelException):
    pass


class EmptyPassword(SelectelException):
    pass


class EmptyContainerName(SelectelException):
    pass



from django import apps
from django.conf import settings
from django.core.checks import Tags, Warning, register

HINT = (
    "Since 1.0, to improve the experience of multiple containers in the "
    "same project using, the settings format has been changed. For now, "
    "you should not use SELECTEL_USERNAME, SELECTEL_PASSWORD, "
    "SELECTEL_CONTAINER_NAME etc. settings. Consider using "
    "a SELECTEL_STORAGE  dictionary instead, like DATABASES or CACHES ones. "
)


class DjangoSelectelStorageAppConfig(apps.AppConfig):
    name = 'django_selectel_storage'


@register(Tags.compatibility)
def settings_compat_check(app_configs, **kwargs):

    if any((
        hasattr(settings, 'SELECTEL_USERNAME'),
        hasattr(settings, 'SELECTEL_PASSWORD'),
        hasattr(settings, 'SELECTEL_CONTAINER_NAME'),
    )):
        return [
            Warning(
                msg='Obsolete config format',
                hint=HINT,
                id='django_selectel_storage.W001',
            )
        ]
    return []



import requests
from django.conf import settings

from .exceptions import (
    EmptyContainerName,
    EmptyPassword,
    EmptyUsername,
    InvalidSchema,
    SelectelException,
)

try:
    from urllib import parse as urlparse   # PY3
except ImportError:
    import urlparse                        # PY2


MAX_RETRIES = 3
POOL_CONNECTIONS = 50
POOL_MAXSIZE = 50

KNOWN_OPTS = (
    'max_retries',
    'pool_conns',
    'pool_maxsize'
)


def extract(opts, key, default=None, cast=None):
    if key not in opts:
        return default
    return cast(opts[key][0]) if cast is not None else opts[key][0]


def empty_username():
    raise EmptyUsername('An username should be given to access the '
                        'Selectel cloud storage')


def empty_password():
    raise EmptyPassword('A password should be given to access the'
                        'Selectel cloud storage')


def empty_container_name():
    raise EmptyContainerName('Looks like both username and password are '
                             'given, but container name seems to be empty')


def parse_dsn(string):
    bits = urlparse.urlparse(string)  # type: urlparse.ParseResult

    if bits.scheme.lower() != 'selectel':
        raise InvalidSchema('The only supported schema here is "selectel://"')

    if not bits.hostname:
        empty_container_name()

    if bits.hostname and bits.netloc.count(':') == 1:
        if bits.password is None and bits.username is None:
            empty_container_name()

        if not bits.password:
            empty_password()

        if not bits.username:
            empty_username()

    if not bits.username:
        empty_username()

    if not bits.password:
        empty_password()

    username = bits.username
    password = bits.password

    custom_domain = bits.hostname
    container = bits.path.strip('/')

    if bits.path in ['', '/']:
        custom_domain = None
        container = bits.hostname

    options = urlparse.parse_qs(bits.query)

    return {
        'USERNAME': username,
        'PASSWORD': password,
        'CONTAINER': container,
        'CUSTOM_DOMAIN': custom_domain,
        'OPTIONS': {
            'MAX_RETRIES': extract(
                opts=options,
                key='max_retries',
                cast=int,
                default=MAX_RETRIES
            ),
            'POOL_CONNECTIONS': extract(
                opts=options,
                key='pool_conns',
                cast=int,
                default=POOL_CONNECTIONS
            ),
            'POOL_MAXSIZE': extract(
                opts=options,
                key='pool_maxsize',
                cast=int,
                default=POOL_MAXSIZE
            ),
            'REQUESTS_FACTORY': extract(
                opts=options,
                key='requests_factory',
                cast=str,
                default='django_selectel_storage.utils.requests_factory'
            )
        }
    }


def read_config(args, kwargs):
    detected_schema = None

    # Case 1: Storage() with no args and kwargs
    if not args and not kwargs:
        detected_schema = 'default'

    # Case 2: the only positional argument: a data source string
    # like Storage("selectel://user:pass@name") or a schema name
    # like Storage("schema-name")
    if len(args) == 1 and not kwargs:
        try:
            return parse_dsn(args[0])
        except SelectelException:
            detected_schema = args[0]

    # Case 3: Using the "default" schema when there is not "dsn=" keyword arg.
    if kwargs.get('dsn') is None and all((
        'container' not in kwargs,
        'username' not in kwargs,
        'password' not in kwargs
    )):
        detected_schema = args[0] if args else 'default'

    # Case 4: Storage(storage='default')   default schema, the same as 1)
    if 'storage' in kwargs:
        detected_schema = kwargs['storage']

    if detected_schema is not None:
        return settings.SELECTEL_STORAGES[detected_schema]

    # 5. Storage(dsn='selectel://user:pass@container')
    if 'dsn' in kwargs:
        return parse_dsn(kwargs['dsn'])

    if all((
        'container' in kwargs,
        'username' in kwargs,
        'password' in kwargs
    )):
        dsn_format = 'selectel://{username}:{password}@{container}'
        if 'custom_domain'in kwargs:
            dsn_format = (
                'selectel://{username}:{password}@{custom_domain}/{container}'
            )
        if 'options' in kwargs:
            dsn_format += '/?'
        return parse_dsn(dsn_format.format(**kwargs))
    # 7. Storage(container='container', user='...', password='', options=None)


def requests_factory(config):
    return requests.Session()

# def get_requests_adapter(self, **kwargs):
#     return requests.adapters.HTTPAdapter(
#         max_retries=setting('SELECTEL_MAX_RETRIES',
#                             MAX_RETRIES),
#         pool_connections=setting('SELECTEL_POOL_CONNECTIONS',
#                                  POOL_CONNECTIONS),
#         pool_maxsize=setting('SELECTEL_POOL_MAXSIZE',
#                              POOL_MAXSIZE))
#
# def setup_requests_adapter(self, **kwargs):
#     adapter = self.get_requests_adapter(**kwargs)
#     self.mount_requests_adapter('http://', adapter)
#     self.mount_requests_adapter('https://', adapter)
#
# def mount_requests_adapter(self, prefix, adapter):
#     self.container.storage.session.mount(prefix, adapter)



from django.core.files import base, storage

from .selectel import Container
from .utils import read_config


class SelectelStorage(storage.Storage):
    def __init__(self, *args, **kwargs):
        self.config = read_config(args, kwargs)
        self.container = Container(self.config)

    def _open(self, name, mode='rb'):
        return base.ContentFile(self.container.open(name).read())

    def _save(self, name, content):
        self.container.save(name, content, metadata=None)
        return name

    def delete(self, name):
        self.container.delete(name)

    def exists(self, name):
        return self.container.exists(name)

    def listdir(self, path):
        dirs, files = set(), set()
        results = self.container.list(path)
        for key, metadata in results.items():
            bits = key[len(path):].lstrip('/').split('/')
            (dirs if len(bits) > 1 else files).add(bits[0])
        return list(dirs), list(files)

    def size(self, name):
        return self.container.size(name)

    def url(self, name):
        if not self.config.get('CUSTOM_DOMAIN'):
            return self.container.build_url(name)
        else:
            custom_domain = self.config['CUSTOM_DOMAIN'].rstrip('/')
            if not custom_domain.lower().startswith(('http://', 'https://')):
                custom_domain = 'https://{0}'.format(custom_domain)
            return '{custom_domain}/{name}'.format(
                custom_domain=custom_domain,
                name=name.strip('/'),
            )

    def save_with_metadata(self, name, content, metadata=None):
        self.container.save(name, content, metadata=metadata)
        return name


class SelectelStaticStorage(SelectelStorage):
    pass



import sys

PY2 = sys.version_info.major == 2
PY3 = sys.version_info.major == 3

try:
    TEXT_TYPE = unicode
except NameError:
    TEXT_TYPE = str


def b(s):
    return s.encode('latin-1') if PY3 else s



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['django_selectel_storage']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.23.0,<3.0.0']

setup_kwargs = {
    'name': 'django-selectel-storage',
    'version': '1.0.1',
    'description': 'A Django storage backend allowing you to easily save user-generated and static files inside Selectel Cloud storage rather than a local filesystem, as Django does by default.',
    'long_description': "=======================\ndjango-selectel-storage\n=======================\n\n\n\n.. image:: https://badge.fury.io/py/django-selectel-storage.svg\n    :target: https://badge.fury.io/py/django-selectel-storage\n\n.. image:: https://img.shields.io/pypi/l/django-selectel-storage\n    :target: https://raw.githubusercontent.com/marazmiki/django-selectel-storage/master/LICENSE\n    :alt: The project license\n\n.. image:: https://travis-ci.org/marazmiki/django-selectel-storage.svg?branch=master\n    :target: https://travis-ci.org/marazmiki/django-selectel-storage\n    :alt: Travis CI build status\n\n.. image:: https://coveralls.io/repos/marazmiki/django-selectel-storage/badge.svg?branch=master\n    :target: https://coveralls.io/r/marazmiki/django-selectel-storage?branch=master\n    :alt: Code coverage percentage\n\n.. image:: https://pypip.in/wheel/django-selectel-storage/badge.svg\n     :target: https://pypi.python.org/pypi/django-selectel-storage/\n     :alt: Wheel Status\n\n.. image:: https://img.shields.io/pypi/pyversions/django-selectel-storage.svg\n     :target: https://img.shields.io/pypi/pyversions/django-selectel-storage.svg\n     :alt: Supported Python versions\n\n.. image:: https://img.shields.io/pypi/djversions/django-selectel-storage.svg\n     :target: https://pypi.org/project/django-selectel-storage/\n     :alt: Supported Django versions\n\n.. image:: https://readthedocs.org/projects/django-selectel-storage/badge/?version=latest\n     :target: https://django-ulogin.readthedocs.io/ru/latest/?badge=latest\n     :alt: Documentation Status\n\n.. image:: https://api.codacy.com/project/badge/Grade/f143275acdf249328a4968b62a94e100\n   :alt: Codacy Badge\n   :target: https://app.codacy.com/manual/marazmiki/django-selectel-storage?utm_source=github.com&utm_medium=referral&utm_content=marazmiki/django-selectel-storage&utm_campaign=Badge_Grade_Dashboard\n\n\nThis application allows you easily save media and static files into Selectel cloud storage.\n\n\nInstallation\n------------\n\n1. Install the package\n\n.. code:: bash\n\n    pip install django-selectel-storage\n\n\n2. Add to your settings module:\n\n.. code:: python\n\n    DEFAULT_FILE_STORAGE = 'django_selectel_storage.storage.SelectelStorage'\n    SELECTEL_STORAGES = {\n        'default': {\n            'USERNAME': 'xxxx_user1',\n            'PASSWORD': 'secret',\n            'CONTAINER_NAME': 'bucket',\n        },\n        'yet-another-schema': {\n            'USERNAME': 'yyyy_user2',\n            'PASSWORD': 'mystery',\n            'CONTAINER_NAME': 'box',\n\n        },\n    }\n\nPlease see details in the `documentation <https://django-selectel-storage.readthedocs.io/en/latest/>`_.\n",
    'author': 'Mikhail Porokhovnichenko',
    'author_email': 'marazmiki@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/marazmiki/django-selectel-storage',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*',
}


setup(**setup_kwargs)
